@@ -32,9 +32,22 @@
 /* 16byte-aligns a buffer for libmpeg2 */
 #define ALIGN_16(p) ((void *)(((uintptr_t)(p) + 15) & ~((uintptr_t)15)))
 
+/* mpeg2dec changed a struct name after 0.3.1, here's a workaround */
+/* mpeg2dec also only defined MPEG2_RELEASE after 0.3.1
+   #if MPEG2_RELEASE < MPEG2_VERSION(0,3,2)
+*/
+#ifndef MPEG2_RELEASE
+#define MPEG2_VERSION(a,b,c) ((((a)&0xff)<<16)|(((b)&0xff)<<8)|((c)&0xff))
+#define MPEG2_RELEASE MPEG2_VERSION(0,3,1)
+typedef picture_t mpeg2_picture_t;
+typedef gint mpeg2_state_t;
+
+#define STATE_BUFFER 0
+#endif
+
 GST_DEBUG_CATEGORY_STATIC (mpeg2dec_debug);
-#define GST_CAT_DEFAULT mpeg2dec_debug
-GST_DEBUG_CATEGORY_STATIC (CAT_PERFORMANCE);
+#define GST_CAT_DEFAULT (mpeg2dec_debug)
+GST_DEBUG_CATEGORY_EXTERN (GST_CAT_PERFORMANCE);
 
 /* Send a warning message about decoding errors after receiving this many
  * STATE_INVALID return values from mpeg2_parse. -1 means never.
@@ -62,8 +75,6 @@ GST_STATIC_PAD_TEMPLATE ("src",
 
 #define gst_mpeg2dec_parent_class parent_class
 G_DEFINE_TYPE (GstMpeg2dec, gst_mpeg2dec, GST_TYPE_VIDEO_DECODER);
-GST_ELEMENT_REGISTER_DEFINE (mpeg2dec, "mpeg2dec", GST_RANK_SECONDARY,
-    GST_TYPE_MPEG2DEC);
 
 static void gst_mpeg2dec_finalize (GObject * object);
 
@@ -82,7 +93,7 @@ static gboolean gst_mpeg2dec_decide_allocation (GstVideoDecoder * decoder,
     GstQuery * query);
 
 static void gst_mpeg2dec_clear_buffers (GstMpeg2dec * mpeg2dec);
-static GstFlowReturn gst_mpeg2dec_crop_buffer (GstMpeg2dec * dec,
+static gboolean gst_mpeg2dec_crop_buffer (GstMpeg2dec * dec,
     GstVideoCodecFrame * in_frame, GstVideoFrame * in_vframe);
 
 static void
@@ -94,10 +105,10 @@ gst_mpeg2dec_class_init (GstMpeg2decClass * klass)
 
   gobject_class->finalize = gst_mpeg2dec_finalize;
 
-  gst_element_class_add_static_pad_template (element_class,
-      &src_template_factory);
-  gst_element_class_add_static_pad_template (element_class,
-      &sink_template_factory);
+  gst_element_class_add_pad_template (element_class,
+      gst_static_pad_template_get (&src_template_factory));
+  gst_element_class_add_pad_template (element_class,
+      gst_static_pad_template_get (&sink_template_factory));
   gst_element_class_set_static_metadata (element_class,
       "mpeg1 and mpeg2 video decoder", "Codec/Decoder/Video",
       "Uses libmpeg2 to decode MPEG video streams",
@@ -117,20 +128,12 @@ gst_mpeg2dec_class_init (GstMpeg2decClass * klass)
 
   GST_DEBUG_CATEGORY_INIT (mpeg2dec_debug, "mpeg2dec", 0,
       "MPEG-2 Video Decoder");
-  GST_DEBUG_CATEGORY_GET (CAT_PERFORMANCE, "GST_PERFORMANCE");
 }
 
 static void
 gst_mpeg2dec_init (GstMpeg2dec * mpeg2dec)
 {
   gst_video_decoder_set_packetized (GST_VIDEO_DECODER (mpeg2dec), TRUE);
-  gst_video_decoder_set_needs_format (GST_VIDEO_DECODER (mpeg2dec), TRUE);
-  gst_video_decoder_set_use_default_pad_acceptcaps (GST_VIDEO_DECODER_CAST
-      (mpeg2dec), TRUE);
-  GST_PAD_SET_ACCEPT_TEMPLATE (GST_VIDEO_DECODER_SINK_PAD (mpeg2dec));
-
-  g_warning_once
-      ("The `mpeg2dec` plugin is deprecated and will be removed in the next release cycle. `avdec_mpeg2video` from gst-libav should be used instead");
 
   /* initialize the mpeg2dec acceleration */
 }
@@ -218,11 +221,6 @@ gst_mpeg2dec_stop (GstVideoDecoder * decoder)
     gst_video_codec_state_unref (mpeg2dec->input_state);
   mpeg2dec->input_state = NULL;
 
-  if (mpeg2dec->downstream_pool) {
-    gst_buffer_pool_set_active (mpeg2dec->downstream_pool, FALSE);
-    gst_object_unref (mpeg2dec->downstream_pool);
-  }
-
   return TRUE;
 }
 
@@ -238,9 +236,6 @@ gst_mpeg2dec_flush (GstVideoDecoder * decoder)
 
   gst_mpeg2dec_clear_buffers (mpeg2dec);
 
-  if (mpeg2dec->downstream_pool)
-    gst_buffer_pool_set_active (mpeg2dec->downstream_pool, FALSE);
-
   return TRUE;
 }
 
@@ -250,52 +245,17 @@ gst_mpeg2dec_finish (GstVideoDecoder * decoder)
   return GST_FLOW_OK;
 }
 
-static GstBufferPool *
-gst_mpeg2dec_create_generic_pool (GstAllocator * allocator,
-    GstAllocationParams * params, GstCaps * caps, guint size, guint min,
-    guint max, GstStructure ** out_config)
-{
-  GstBufferPool *pool;
-  GstStructure *config;
-
-  pool = gst_video_buffer_pool_new ();
-  {
-    gchar *name = g_strdup_printf ("mpeg2dec-%p-pool", pool);
-    g_object_set (pool, "name", name, NULL);
-    g_free (name);
-  }
-  config = gst_buffer_pool_get_config (pool);
-
-  gst_buffer_pool_config_set_allocator (config, allocator, params);
-  gst_buffer_pool_config_set_params (config, caps, size, min, max);
-  gst_buffer_pool_config_add_option (config, GST_BUFFER_POOL_OPTION_VIDEO_META);
-
-  *out_config = config;
-  return pool;
-}
-
 static gboolean
 gst_mpeg2dec_decide_allocation (GstVideoDecoder * decoder, GstQuery * query)
 {
   GstMpeg2dec *dec = GST_MPEG2DEC (decoder);
+  GstVideoCodecState *state;
   GstBufferPool *pool;
   guint size, min, max;
-  GstStructure *config, *down_config = NULL;
+  GstStructure *config;
   GstAllocator *allocator;
   GstAllocationParams params;
   gboolean update_allocator;
-  gboolean has_videometa = FALSE;
-  GstCaps *caps;
-
-  /* Get rid of ancient pool */
-  if (dec->downstream_pool) {
-    gst_buffer_pool_set_active (dec->downstream_pool, FALSE);
-    gst_object_unref (dec->downstream_pool);
-    dec->downstream_pool = NULL;
-  }
-
-  /* Get negotiated allocation caps */
-  gst_query_parse_allocation (query, &caps, NULL);
 
   /* Set allocation parameters to guarantee 16-byte aligned output buffers */
   if (gst_query_get_n_allocation_params (query) > 0) {
@@ -313,174 +273,47 @@ gst_mpeg2dec_decide_allocation (GstVideoDecoder * decoder, GstQuery * query)
     gst_query_set_nth_allocation_param (query, 0, allocator, &params);
   else
     gst_query_add_allocation_param (query, allocator, &params);
+  if (allocator)
+    gst_object_unref (allocator);
 
   /* Now chain up to the parent class to guarantee that we can
    * get a buffer pool from the query */
   if (!GST_VIDEO_DECODER_CLASS (parent_class)->decide_allocation (decoder,
-          query)) {
-    if (allocator)
-      gst_object_unref (allocator);
+          query))
     return FALSE;
-  }
+
+  state = gst_video_decoder_get_output_state (decoder);
 
   gst_query_parse_nth_allocation_pool (query, 0, &pool, &size, &min, &max);
 
+  dec->has_cropping = FALSE;
   config = gst_buffer_pool_get_config (pool);
   if (gst_query_find_allocation_meta (query, GST_VIDEO_META_API_TYPE, NULL)) {
     gst_buffer_pool_config_add_option (config,
         GST_BUFFER_POOL_OPTION_VIDEO_META);
-    has_videometa = TRUE;
+    dec->has_cropping =
+        gst_query_find_allocation_meta (query, GST_VIDEO_CROP_META_API_TYPE,
+        NULL);
   }
 
-  if (dec->need_alignment) {
-    /* If downstream does not support video meta, we will have to copy, keep
-     * the downstream pool to avoid double copying */
-    if (!has_videometa) {
-      dec->downstream_pool = pool;
-      pool = NULL;
-      down_config = config;
-      config = NULL;
-      min = 2;
-      max = 0;
-    }
+  if (dec->has_cropping) {
+    GstCaps *caps;
 
-    /* In case downstream support video meta, but the downstream pool does not
-     * have alignment support, discard downstream pool and use video pool */
-    else if (!gst_buffer_pool_has_option (pool,
-            GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT)) {
-      gst_object_unref (pool);
-      pool = NULL;
-      gst_structure_free (config);
-      config = NULL;
-    }
-
-    if (!pool)
-      pool = gst_mpeg2dec_create_generic_pool (allocator, &params, caps, size,
-          min, max, &config);
-
-    gst_buffer_pool_config_add_option (config,
-        GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT);
-    gst_buffer_pool_config_set_video_alignment (config, &dec->valign);
+    /* Calculate uncropped size */
+    size = MAX (size, dec->decoded_info.size);
+    caps = gst_video_info_to_caps (&dec->decoded_info);
+    gst_buffer_pool_config_set_params (config, caps, size, min, max);
+    gst_caps_unref (caps);
   }
 
-  if (allocator)
-    gst_object_unref (allocator);
-
-  /* If we are copying out, we'll need to setup and activate the other pool */
-  if (dec->downstream_pool) {
-    if (!gst_buffer_pool_set_config (dec->downstream_pool, down_config)) {
-      down_config = gst_buffer_pool_get_config (dec->downstream_pool);
-      if (!gst_buffer_pool_config_validate_params (down_config, caps, size, min,
-              max)) {
-        gst_structure_free (down_config);
-        goto config_failed;
-      }
-
-      if (!gst_buffer_pool_set_config (dec->downstream_pool, down_config))
-        goto config_failed;
-    }
-
-    if (!gst_buffer_pool_set_active (dec->downstream_pool, TRUE))
-      goto activate_failed;
-  }
-
-  /* Now configure the pool, if the pool had made some changes, it will
-   * return FALSE. Validate the changes ...*/
-  if (!gst_buffer_pool_set_config (pool, config)) {
-    config = gst_buffer_pool_get_config (pool);
-
-    /* Check basic params */
-    if (!gst_buffer_pool_config_validate_params (config, caps, size, min, max)) {
-      gst_structure_free (config);
-      goto config_failed;
-    }
-
-    /* If needed, check that resulting alignment is still valid */
-    if (dec->need_alignment) {
-      GstVideoAlignment valign;
-
-      if (!gst_buffer_pool_config_get_video_alignment (config, &valign)) {
-        gst_structure_free (config);
-        goto config_failed;
-      }
-
-      if (valign.padding_left != 0 || valign.padding_top != 0
-          || valign.padding_right < dec->valign.padding_right
-          || valign.padding_bottom < dec->valign.padding_bottom) {
-        gst_structure_free (config);
-        goto config_failed;
-      }
-    }
-
-    if (!gst_buffer_pool_set_config (pool, config))
-      goto config_failed;
-  }
-
-  /* For external pools, we need to check strides */
-  if (!GST_IS_VIDEO_BUFFER_POOL (pool) && has_videometa) {
-    GstBuffer *buffer;
-    const GstVideoFormatInfo *finfo;
-    GstVideoMeta *vmeta;
-    gint uv_stride;
-
-    if (!gst_buffer_pool_set_active (pool, TRUE))
-      goto activate_failed;
-
-    if (gst_buffer_pool_acquire_buffer (pool, &buffer, NULL) != GST_FLOW_OK) {
-      gst_buffer_pool_set_active (pool, FALSE);
-      goto acquire_failed;
-    }
-
-    vmeta = gst_buffer_get_video_meta (buffer);
-    finfo = gst_video_format_get_info (vmeta->format);
-
-    /* Check that strides are compatible. In this case, we can scale the
-     * stride directly since all the pixel strides for the formats we support
-     * is 1 */
-    uv_stride = GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (finfo, 1, vmeta->stride[0]);
-    if (uv_stride != vmeta->stride[1] || uv_stride != vmeta->stride[2]) {
-      gst_buffer_pool_set_active (pool, FALSE);
-      gst_object_unref (pool);
-
-      pool = gst_mpeg2dec_create_generic_pool (allocator, &params, caps, size,
-          min, max, &config);
-
-      if (dec->need_alignment) {
-        gst_buffer_pool_config_add_option (config,
-            GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT);
-        gst_buffer_pool_config_set_video_alignment (config, &dec->valign);
-      }
-
-      /* Generic pool don't fail on _set_config() */
-      gst_buffer_pool_set_config (pool, config);
-    }
-
-    gst_buffer_unref (buffer);
-  }
+  gst_buffer_pool_set_config (pool, config);
 
   gst_query_set_nth_allocation_pool (query, 0, pool, size, min, max);
+
   gst_object_unref (pool);
+  gst_video_codec_state_unref (state);
 
   return TRUE;
-
-config_failed:
-  gst_object_unref (pool);
-  GST_ELEMENT_ERROR (dec, RESOURCE, SETTINGS,
-      ("Failed to configure buffer pool"),
-      ("Configuration is most likely invalid, please report this issue."));
-  return FALSE;
-
-activate_failed:
-  gst_object_unref (pool);
-  GST_ELEMENT_ERROR (dec, RESOURCE, SETTINGS,
-      ("Failed to activate buffer pool"), (NULL));
-  return FALSE;
-
-acquire_failed:
-  gst_object_unref (pool);
-  GST_ELEMENT_ERROR (dec, RESOURCE, SETTINGS,
-      ("Failed to acquire a buffer"), (NULL));
-  return FALSE;
 }
 
 static GstFlowReturn
@@ -490,32 +323,52 @@ gst_mpeg2dec_crop_buffer (GstMpeg2dec * dec, GstVideoCodecFrame * in_frame,
   GstVideoCodecState *state;
   GstVideoInfo *info;
   GstVideoInfo *dinfo;
+  guint c, n_planes;
   GstVideoFrame output_frame;
   GstFlowReturn ret;
-  GstBuffer *buffer = NULL;
 
   state = gst_video_decoder_get_output_state (GST_VIDEO_DECODER (dec));
   info = &state->info;
   dinfo = &dec->decoded_info;
 
-  GST_CAT_LOG_OBJECT (CAT_PERFORMANCE, dec,
+  GST_CAT_LOG_OBJECT (GST_CAT_PERFORMANCE, dec,
       "Copying input buffer %ux%u (%" G_GSIZE_FORMAT ") to output buffer "
       "%ux%u (%" G_GSIZE_FORMAT ")", dinfo->width, dinfo->height,
       dinfo->size, info->width, info->height, info->size);
 
-  ret = gst_buffer_pool_acquire_buffer (dec->downstream_pool, &buffer, NULL);
+  ret =
+      gst_video_decoder_allocate_output_frame (GST_VIDEO_DECODER (dec),
+      in_frame);
   if (ret != GST_FLOW_OK)
     goto beach;
 
-  if (!gst_video_frame_map (&output_frame, info, buffer, GST_MAP_WRITE))
+  if (!gst_video_frame_map (&output_frame, info, in_frame->output_buffer,
+          GST_MAP_WRITE))
     goto map_fail;
 
-  if (in_frame->output_buffer)
-    gst_buffer_unref (in_frame->output_buffer);
-  in_frame->output_buffer = buffer;
+  n_planes = GST_VIDEO_FRAME_N_PLANES (&output_frame);
+  for (c = 0; c < n_planes; c++) {
+    guint w, h, j;
+    guint8 *sp, *dp;
+    gint ss, ds;
 
-  if (!gst_video_frame_copy (&output_frame, input_vframe))
-    goto copy_failed;
+    sp = GST_VIDEO_FRAME_PLANE_DATA (input_vframe, c);
+    dp = GST_VIDEO_FRAME_PLANE_DATA (&output_frame, c);
+
+    ss = GST_VIDEO_FRAME_PLANE_STRIDE (input_vframe, c);
+    ds = GST_VIDEO_FRAME_PLANE_STRIDE (&output_frame, c);
+
+    w = MIN (ABS (ss), ABS (ds));
+    h = GST_VIDEO_FRAME_COMP_HEIGHT (&output_frame, c);
+
+    GST_CAT_DEBUG (GST_CAT_PERFORMANCE, "copy plane %u, w:%u h:%u ", c, w, h);
+
+    for (j = 0; j < h; j++) {
+      memcpy (dp, sp, w);
+      dp += ds;
+      sp += ss;
+    }
+  }
 
   gst_video_frame_unmap (&output_frame);
 
@@ -533,14 +386,41 @@ map_fail:
     gst_video_codec_state_unref (state);
     return GST_FLOW_ERROR;
   }
+}
 
-copy_failed:
-  {
-    GST_ERROR_OBJECT (dec, "Failed to copy output frame");
-    gst_video_frame_unmap (&output_frame);
-    gst_video_codec_state_unref (state);
-    return GST_FLOW_ERROR;
+static void
+frame_user_data_destroy_notify (GstBuffer * buf)
+{
+  GST_DEBUG ("Releasing buffer %p", buf);
+  if (buf)
+    gst_buffer_unref (buf);
+}
+
+static GstFlowReturn
+gst_mpeg2dec_alloc_sized_buf (GstMpeg2dec * mpeg2dec, guint size,
+    GstVideoCodecFrame * frame, GstBuffer ** buffer)
+{
+  GstFlowReturn ret = GST_FLOW_OK;
+  GstVideoCodecState *state;
+
+  state = gst_video_decoder_get_output_state (GST_VIDEO_DECODER (mpeg2dec));
+
+  if (!mpeg2dec->need_cropping || mpeg2dec->has_cropping) {
+    ret =
+        gst_video_decoder_allocate_output_frame (GST_VIDEO_DECODER (mpeg2dec),
+        frame);
+    *buffer = frame->output_buffer;
+  } else {
+    GstAllocationParams params = { 0, 15, 0, 0 };
+
+    *buffer = gst_buffer_new_allocate (NULL, size, &params);
+    gst_video_codec_frame_set_user_data (frame, *buffer,
+        (GDestroyNotify) frame_user_data_destroy_notify);
   }
+
+  gst_video_codec_state_unref (state);
+
+  return ret;
 }
 
 typedef struct
@@ -556,7 +436,7 @@ gst_mpeg2dec_clear_buffers (GstMpeg2dec * mpeg2dec)
   while ((l = g_list_first (mpeg2dec->buffers))) {
     GstMpeg2DecBuffer *mbuf = l->data;
     gst_video_frame_unmap (&mbuf->frame);
-    g_free (mbuf);
+    g_slice_free (GstMpeg2DecBuffer, mbuf);
     mpeg2dec->buffers = g_list_delete_link (mpeg2dec->buffers, l);
   }
 }
@@ -569,7 +449,7 @@ gst_mpeg2dec_save_buffer (GstMpeg2dec * mpeg2dec, gint id,
 
   GST_LOG_OBJECT (mpeg2dec, "Saving local info for frame %d", id);
 
-  mbuf = g_new0 (GstMpeg2DecBuffer, 1);
+  mbuf = g_slice_new0 (GstMpeg2DecBuffer);
   mbuf->id = id;
   mbuf->frame = *frame;
 
@@ -593,7 +473,7 @@ gst_mpeg2dec_discard_buffer (GstMpeg2dec * mpeg2dec, gint id)
   if (l) {
     GstMpeg2DecBuffer *mbuf = l->data;
     gst_video_frame_unmap (&mbuf->frame);
-    g_free (mbuf);
+    g_slice_free (GstMpeg2DecBuffer, mbuf);
     mpeg2dec->buffers = g_list_delete_link (mpeg2dec->buffers, l);
     GST_LOG_OBJECT (mpeg2dec, "Discarded local info for frame %d", id);
   } else {
@@ -616,6 +496,67 @@ gst_mpeg2dec_get_buffer (GstMpeg2dec * mpeg2dec, gint id)
   return NULL;
 }
 
+static GstFlowReturn
+gst_mpeg2dec_alloc_buffer (GstMpeg2dec * mpeg2dec, GstVideoCodecFrame * frame,
+    GstBuffer ** buffer)
+{
+  GstFlowReturn ret;
+  GstVideoFrame vframe;
+  guint8 *buf[3];
+
+  ret =
+      gst_mpeg2dec_alloc_sized_buf (mpeg2dec, mpeg2dec->decoded_info.size,
+      frame, buffer);
+  if (G_UNLIKELY (ret != GST_FLOW_OK))
+    goto beach;
+
+  if (mpeg2dec->need_cropping && mpeg2dec->has_cropping) {
+    GstVideoCropMeta *crop;
+    GstVideoCodecState *state;
+    GstVideoInfo *vinfo;
+
+    state = gst_video_decoder_get_output_state (GST_VIDEO_DECODER (mpeg2dec));
+    vinfo = &state->info;
+
+    crop = gst_buffer_add_video_crop_meta (frame->output_buffer);
+    /* we can do things slightly more efficient when we know that
+     * downstream understands clipping */
+    crop->x = 0;
+    crop->y = 0;
+    crop->width = vinfo->width;
+    crop->height = vinfo->height;
+
+    gst_video_codec_state_unref (state);
+  }
+
+  if (!gst_video_frame_map (&vframe, &mpeg2dec->decoded_info, *buffer,
+          GST_MAP_READ | GST_MAP_WRITE))
+    goto map_fail;
+
+  buf[0] = GST_VIDEO_FRAME_PLANE_DATA (&vframe, 0);
+  buf[1] = GST_VIDEO_FRAME_PLANE_DATA (&vframe, 1);
+  buf[2] = GST_VIDEO_FRAME_PLANE_DATA (&vframe, 2);
+
+  GST_DEBUG_OBJECT (mpeg2dec, "set_buf: %p %p %p, frame %i",
+      buf[0], buf[1], buf[2], frame->system_frame_number);
+
+  /* Note: We use a non-null 'id' value to make the distinction
+   * between the dummy buffers (which have an id of NULL) and the
+   * ones we did */
+  mpeg2_set_buf (mpeg2dec->decoder, buf,
+      GINT_TO_POINTER (frame->system_frame_number + 1));
+  gst_mpeg2dec_save_buffer (mpeg2dec, frame->system_frame_number, &vframe);
+
+beach:
+  return ret;
+
+map_fail:
+  {
+    GST_ERROR_OBJECT (mpeg2dec, "Failed to map frame");
+    return GST_FLOW_ERROR;
+  }
+}
+
 static void
 init_dummybuf (GstMpeg2dec * mpeg2dec)
 {
@@ -639,6 +580,7 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   GstClockTime latency;
   const mpeg2_sequence_t *sequence;
   GstVideoCodecState *state;
+  GstVideoInfo *dinfo = &mpeg2dec->decoded_info;
   GstVideoInfo *vinfo;
   GstVideoFormat format;
 
@@ -658,21 +600,13 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
       sequence->picture_width, sequence->picture_height, sequence->width,
       sequence->height);
 
-  gst_video_alignment_reset (&mpeg2dec->valign);
-
-  if (sequence->picture_width < sequence->width ||
-      sequence->picture_height < sequence->height) {
+  if (sequence->picture_width != sequence->width ||
+      sequence->picture_height != sequence->height) {
     GST_DEBUG_OBJECT (mpeg2dec, "we need to crop");
-    mpeg2dec->valign.padding_right = sequence->width - sequence->picture_width;
-    mpeg2dec->valign.padding_bottom =
-        sequence->height - sequence->picture_height;
-    mpeg2dec->need_alignment = TRUE;
-  } else if (sequence->picture_width == sequence->width ||
-      sequence->picture_height == sequence->height) {
-    GST_DEBUG_OBJECT (mpeg2dec, "no cropping needed");
-    mpeg2dec->need_alignment = FALSE;
+    mpeg2dec->need_cropping = TRUE;
   } else {
-    goto invalid_picture;
+    GST_DEBUG_OBJECT (mpeg2dec, "no cropping needed");
+    mpeg2dec->need_cropping = FALSE;
   }
 
   /* get subsampling */
@@ -698,12 +632,16 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   if (GST_VIDEO_INFO_PAR_N (vinfo) == 1 &&
       GST_VIDEO_INFO_PAR_D (vinfo) == 1 &&
       sequence->pixel_width != 0 && sequence->pixel_height != 0) {
+#if MPEG2_RELEASE >= MPEG2_VERSION(0,5,0)
     guint pixel_width, pixel_height;
-
     if (mpeg2_guess_aspect (sequence, &pixel_width, &pixel_height)) {
       vinfo->par_n = pixel_width;
       vinfo->par_d = pixel_height;
     }
+#else
+    vinfo->par_n = sequence->pixel_width;
+    vinfo->par_d = sequence->pixel_height;
+#endif
     GST_DEBUG_OBJECT (mpeg2dec, "Setting PAR %d x %d",
         vinfo->par_n, vinfo->par_d);
   }
@@ -739,7 +677,7 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
         /* 0 forbidden */
         /* 2 unspecified */
         /* 3 reserved */
-        /* 8-255 reserved */
+        /* 8-255 reseved */
       default:
         vinfo->colorimetry.primaries = GST_VIDEO_COLOR_PRIMARIES_UNKNOWN;
         break;
@@ -762,7 +700,7 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
         /* 0 forbidden */
         /* 2 unspecified */
         /* 3 reserved */
-        /* 8-255 reserved */
+        /* 8-255 reseved */
       default:
         vinfo->colorimetry.matrix = GST_VIDEO_COLOR_MATRIX_UNKNOWN;
         break;
@@ -790,7 +728,7 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
         /* 0 forbidden */
         /* 2 unspecified */
         /* 3 reserved */
-        /* 9-255 reserved */
+        /* 9-255 reseved */
       default:
         vinfo->colorimetry.transfer = GST_VIDEO_TRANSFER_UNKNOWN;
         break;
@@ -812,13 +750,13 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
       sequence->flags & SEQ_FLAG_LOW_DELAY,
       sequence->flags & SEQ_FLAG_COLOUR_DESCRIPTION);
 
-  /* Save the padded video information */
-  mpeg2dec->decoded_info = *vinfo;
-  gst_video_info_align (&mpeg2dec->decoded_info, &mpeg2dec->valign);
+  /* we store the codec size before cropping */
+  *dinfo = *vinfo;
+  gst_video_info_set_format (dinfo, format, sequence->width, sequence->height);
 
   /* Mpeg2dec has 2 frame latency to produce a picture and 1 frame latency in
    * it's parser */
-  latency = gst_util_uint64_scale (3 * GST_SECOND, vinfo->fps_d, vinfo->fps_n);
+  latency = gst_util_uint64_scale (3, vinfo->fps_d, vinfo->fps_n);
   gst_video_decoder_set_latency (GST_VIDEO_DECODER (mpeg2dec), latency,
       latency);
 
@@ -852,19 +790,9 @@ invalid_size:
     return GST_FLOW_ERROR;
   }
 
-invalid_picture:
-  {
-    GST_ERROR_OBJECT (mpeg2dec, "Picture dimension bigger then frame: "
-        "%d x %d is bigger then %d x %d", sequence->picture_width,
-        sequence->picture_height, sequence->width, sequence->height);
-    return GST_FLOW_ERROR;
-  }
-
-
 negotiation_fail:
   {
     GST_WARNING_OBJECT (mpeg2dec, "Failed to negotiate with downstream");
-    gst_video_codec_state_unref (state);
     return GST_FLOW_ERROR;
   }
 }
@@ -873,16 +801,14 @@ static GstFlowReturn
 handle_picture (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info,
     GstVideoCodecFrame * frame)
 {
-  GstVideoDecoder *decoder = (GstVideoDecoder *) mpeg2dec;
   GstFlowReturn ret;
   gint type;
   const gchar *type_str = NULL;
   gboolean key_frame = FALSE;
   const mpeg2_picture_t *picture = info->current_picture;
-  GstVideoFrame vframe;
-  guint8 *buf[3];
+  GstBuffer *buffer;
 
-  ret = gst_video_decoder_allocate_output_frame (decoder, frame);
+  ret = gst_mpeg2dec_alloc_buffer (mpeg2dec, frame, &buffer);
   if (ret != GST_FLOW_OK)
     return ret;
 
@@ -914,15 +840,17 @@ handle_picture (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info,
   if (GST_VIDEO_INFO_IS_INTERLACED (&mpeg2dec->decoded_info)) {
     /* This implies SEQ_FLAG_PROGRESSIVE_SEQUENCE is not set */
     if (picture->flags & PIC_FLAG_TOP_FIELD_FIRST) {
-      GST_BUFFER_FLAG_SET (frame->output_buffer, GST_VIDEO_BUFFER_FLAG_TFF);
+      GST_BUFFER_FLAG_SET (buffer, GST_VIDEO_BUFFER_FLAG_TFF);
     }
     if (!(picture->flags & PIC_FLAG_PROGRESSIVE_FRAME)) {
-      GST_BUFFER_FLAG_SET (frame->output_buffer,
-          GST_VIDEO_BUFFER_FLAG_INTERLACED);
+      GST_BUFFER_FLAG_SET (buffer, GST_VIDEO_BUFFER_FLAG_INTERLACED);
     }
+#if MPEG2_RELEASE >= MPEG2_VERSION(0,5,0)
+    /* repeat field introduced in 0.5.0 */
     if (picture->flags & PIC_FLAG_REPEAT_FIRST_FIELD) {
-      GST_BUFFER_FLAG_SET (frame->output_buffer, GST_VIDEO_BUFFER_FLAG_RFF);
+      GST_BUFFER_FLAG_SET (buffer, GST_VIDEO_BUFFER_FLAG_RFF);
     }
+#endif
   }
 
   if (mpeg2dec->discont_state == MPEG2DEC_DISC_NEW_PICTURE && key_frame) {
@@ -934,38 +862,16 @@ handle_picture (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info,
       GST_TIME_FORMAT,
       (picture->flags & PIC_FLAG_PROGRESSIVE_FRAME ? "prog" : "    "),
       (picture->flags & PIC_FLAG_TOP_FIELD_FIRST ? "tff" : "   "),
+#if MPEG2_RELEASE >= MPEG2_VERSION(0,5,0)
       (picture->flags & PIC_FLAG_REPEAT_FIRST_FIELD ? "rff" : "   "),
+#else
+      "unknown rff",
+#endif
       (picture->flags & PIC_FLAG_SKIP ? "skip" : "    "),
       (picture->flags & PIC_FLAG_COMPOSITE_DISPLAY ? "composite" : "         "),
       picture->nb_fields, GST_TIME_ARGS (frame->pts));
 
-  if (!gst_video_frame_map (&vframe, &mpeg2dec->decoded_info,
-          frame->output_buffer, GST_MAP_READ | GST_MAP_WRITE))
-    goto map_fail;
-
-  buf[0] = GST_VIDEO_FRAME_PLANE_DATA (&vframe, 0);
-  buf[1] = GST_VIDEO_FRAME_PLANE_DATA (&vframe, 1);
-  buf[2] = GST_VIDEO_FRAME_PLANE_DATA (&vframe, 2);
-
-  GST_DEBUG_OBJECT (mpeg2dec, "set_buf: %p %p %p, frame %i",
-      buf[0], buf[1], buf[2], frame->system_frame_number);
-
-  /* Note: We use a non-null 'id' value to make the distinction
-   * between the dummy buffers (which have an id of NULL) and the
-   * ones we did */
-  mpeg2_stride (mpeg2dec->decoder, vframe.info.stride[0]);
-  mpeg2_set_buf (mpeg2dec->decoder, buf,
-      GINT_TO_POINTER (frame->system_frame_number + 1));
-  gst_mpeg2dec_save_buffer (mpeg2dec, frame->system_frame_number, &vframe);
-
   return ret;
-
-map_fail:
-  {
-    GST_ELEMENT_ERROR (mpeg2dec, RESOURCE, WRITE, ("Failed to map frame"),
-        (NULL));
-    return GST_FLOW_ERROR;
-  }
 }
 
 static GstFlowReturn
@@ -975,8 +881,7 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   GstVideoCodecFrame *frame;
   const mpeg2_picture_t *picture;
   gboolean key_frame = FALSE;
-  gboolean bidirect_frame = FALSE;
-  gboolean closed_gop = FALSE;
+  GstVideoCodecState *state;
 
   GST_DEBUG_OBJECT (mpeg2dec,
       "fbuf:%p display_picture:%p current_picture:%p fbuf->id:%d",
@@ -991,9 +896,6 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
     goto no_frame;
   picture = info->display_picture;
   key_frame = (picture->flags & PIC_MASK_CODING_TYPE) == PIC_FLAG_CODING_TYPE_I;
-  bidirect_frame =
-      (picture->flags & PIC_MASK_CODING_TYPE) == PIC_FLAG_CODING_TYPE_B;
-  closed_gop = (info->gop->flags & GOP_FLAG_CLOSED_GOP);
 
   GST_DEBUG_OBJECT (mpeg2dec, "picture flags: %d, type: %d, keyframe: %d",
       picture->flags, picture->flags & PIC_MASK_CODING_TYPE, key_frame);
@@ -1012,40 +914,36 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
     return ret;
   }
 
-  /* Skip B-frames if GOP is not closed and waiting for the first keyframe. */
   if (mpeg2dec->discont_state != MPEG2DEC_DISC_NONE) {
-    if (bidirect_frame && !closed_gop) {
-      GST_DEBUG_OBJECT (mpeg2dec, "dropping buffer, discont state %d",
-          mpeg2dec->discont_state);
-      ret = gst_video_decoder_drop_frame (GST_VIDEO_DECODER (mpeg2dec), frame);
-      return ret;
-    }
+    GST_DEBUG_OBJECT (mpeg2dec, "dropping buffer, discont state %d",
+        mpeg2dec->discont_state);
+    ret = gst_video_decoder_drop_frame (GST_VIDEO_DECODER (mpeg2dec), frame);
+    return ret;
   }
 
+  state = gst_video_decoder_get_output_state (GST_VIDEO_DECODER (mpeg2dec));
+
   /* do cropping if the target region is smaller than the input one */
-  if (mpeg2dec->downstream_pool) {
+  if (mpeg2dec->need_cropping && !mpeg2dec->has_cropping) {
     GstVideoFrame *vframe;
 
     if (gst_video_decoder_get_max_decode_time (GST_VIDEO_DECODER (mpeg2dec),
             frame) < 0) {
       GST_DEBUG_OBJECT (mpeg2dec, "dropping buffer crop, too late");
-      return gst_video_decoder_drop_frame (GST_VIDEO_DECODER (mpeg2dec), frame);
+      ret = gst_video_decoder_drop_frame (GST_VIDEO_DECODER (mpeg2dec), frame);
+      goto beach;
     }
 
-    GST_DEBUG_OBJECT (mpeg2dec, "Doing a crop copy of the decoded buffer");
-
+    GST_DEBUG_OBJECT (mpeg2dec, "cropping buffer");
     vframe = gst_mpeg2dec_get_buffer (mpeg2dec, frame->system_frame_number);
     g_assert (vframe != NULL);
     ret = gst_mpeg2dec_crop_buffer (mpeg2dec, frame, vframe);
-
-    if (ret != GST_FLOW_OK) {
-      gst_video_decoder_drop_frame (GST_VIDEO_DECODER (mpeg2dec), frame);
-      return ret;
-    }
   }
 
   ret = gst_video_decoder_finish_frame (GST_VIDEO_DECODER (mpeg2dec), frame);
 
+beach:
+  gst_video_codec_state_unref (state);
   return ret;
 
 no_frame:
@@ -1075,7 +973,6 @@ gst_mpeg2dec_handle_frame (GstVideoDecoder * decoder,
   gst_buffer_ref (buf);
   if (!gst_buffer_map (buf, &minfo, GST_MAP_READ)) {
     GST_ERROR_OBJECT (mpeg2dec, "Failed to map input buffer");
-    gst_buffer_unref (buf);
     return GST_FLOW_ERROR;
   }
 
@@ -1091,11 +988,13 @@ gst_mpeg2dec_handle_frame (GstVideoDecoder * decoder,
     GST_DEBUG_OBJECT (mpeg2dec, "parse state %d", state);
 
     switch (state) {
+#if MPEG2_RELEASE >= MPEG2_VERSION (0, 5, 0)
       case STATE_SEQUENCE_MODIFIED:
         GST_DEBUG_OBJECT (mpeg2dec, "sequence modified");
         mpeg2dec->discont_state = MPEG2DEC_DISC_NEW_PICTURE;
         gst_mpeg2dec_clear_buffers (mpeg2dec);
         /* fall through */
+#endif
       case STATE_SEQUENCE:
         ret = handle_sequence (mpeg2dec, info);
         /* if there is an error handling the sequence
@@ -1126,12 +1025,12 @@ gst_mpeg2dec_handle_frame (GstVideoDecoder * decoder,
         GST_LOG_OBJECT (mpeg2dec,
             "Second picture header encountered. Decoding 2nd field");
         break;
+#if MPEG2_RELEASE >= MPEG2_VERSION (0, 4, 0)
       case STATE_INVALID_END:
         GST_DEBUG_OBJECT (mpeg2dec, "invalid end");
-        /* FALLTHROUGH */
+#endif
       case STATE_END:
         GST_DEBUG_OBJECT (mpeg2dec, "end");
-        /* FALLTHROUGH */
       case STATE_SLICE:
         GST_DEBUG_OBJECT (mpeg2dec, "display_fbuf:%p, discard_fbuf:%p",
             info->display_fbuf, info->discard_fbuf);
@@ -1180,9 +1079,11 @@ done:
 static gboolean
 plugin_init (GstPlugin * plugin)
 {
-  gst_plugin_add_status_warning (plugin,
-      "The `mpeg2dec` plugin is deprecated and will be removed in the next release cycle. `avdec_mpeg2video` from gst-libav should be used instead");
-  return GST_ELEMENT_REGISTER (mpeg2dec, plugin);
+  if (!gst_element_register (plugin, "mpeg2dec", GST_RANK_PRIMARY,
+          GST_TYPE_MPEG2DEC))
+    return FALSE;
+
+  return TRUE;
 }
 
 GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
